mod inbox;
mod users;
mod webfinger;

use std::{
    fmt,
    iter::once,
    panic,
    task::{Context, Poll},
};

use axum::{
    Router,
    http::{StatusCode, Uri},
    routing::{get, post},
};
use inbox::inbox_post;
use log::{error, info, trace};
use tower_layer::Layer;
use tower_service::Service;
use users::users_get;
use webfinger::webfinger_get;

const HOST: &'static str = "testing4host.com";
const USER: &'static str = "test71";

fn init_logger() {
    env_logger::builder()
        .filter_level(log::LevelFilter::Info)
        .parse_default_env()
        .init()
}

#[tokio::main]
async fn main() {
    init_logger();

    panic::set_hook(Box::new(|i| {
        if let Some(loc) = i.location() {
            if let Some(s) = i.payload().downcast_ref::<&str>() {
                error!(
                    "Panicked at {}:{}:{}\n{}",
                    loc.file(),
                    loc.line(),
                    loc.column(),
                    sanitize_printable(s)
                );
            } else if let Some(s) = i.payload().downcast_ref::<String>() {
                error!(
                    "Panicked at {}:{}:{}\n{}",
                    loc.file(),
                    loc.line(),
                    loc.column(),
                    sanitize_printable(s)
                );
            }
        } else {
            if let Some(s) = i.payload().downcast_ref::<&str>() {
                error!("Panicked\n{s:?}",);
            } else if let Some(s) = i.payload().downcast_ref::<String>() {
                error!("Panicked\n{s:?}",);
            }
        }
    }));

    let app = Router::new()
        .route("/", get(root))
        .route("/.well-known/webfinger", get(webfinger_get))
        .route("/users/{users}/", get(users_get))
        .route("/users/{users}", get(users_get))
        .route("/inbox/{users}/", post(inbox_post))
        .fallback(fallback)
        .layer(TraceLayer);

    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn root() -> &'static str {
    "Hello World!"
}

#[derive(Clone)]
pub struct TraceLayer;

async fn fallback(uri: Uri) -> (StatusCode, String) {
    info!("Unknown url! {}", uri);
    (StatusCode::NOT_FOUND, format!("No route for {uri}"))
}

impl<S> Layer<S> for TraceLayer {
    type Service = TraceService<S>;

    fn layer(&self, service: S) -> Self::Service {
        TraceService { service }
    }
}

#[derive(Clone)]
pub struct TraceService<S> {
    service: S,
}

impl<S, Request> Service<Request> for TraceService<S>
where
    S: Service<Request>,
    Request: fmt::Debug,
{
    type Response = S::Response;
    type Error = S::Error;
    type Future = S::Future;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.service.poll_ready(cx)
    }

    fn call(&mut self, request: Request) -> Self::Future {
        trace!("request = {:?}", request);
        self.service.call(request)
    }
}

fn sanitize_printable(src: &str) -> String {
    src.chars()
        .flat_map(|c| -> Box<dyn Iterator<Item = char>> {
            if c.is_control() && !c.is_whitespace() {
                Box::new(c.escape_unicode())
            } else {
                Box::new(once(c))
            }
        })
        .collect()
}
