use std::time::SystemTime;

use base64::prelude::*;
use http::{HeaderMap, HttpSignature, Method, StatusCode, Target};
use httpdate::fmt_http_date;
use key::Key;
use log::debug;
use sha2::Digest;

pub mod http;
pub mod key;

fn sha256sum(bytes: &[u8]) -> String {
    let mut hasher = sha2::Sha256::new();

    hasher.update(bytes);

    return BASE64_STANDARD.encode(hasher.finalize());
}

pub fn sign_request(
    key: Key,
    Target { host, path }: Target,
    method: Method,
    body: &[u8],
) -> HttpSignature {
    let date = fmt_http_date(SystemTime::now());

    let digest = format!("SHA-256={}", sha256sum(body));
    debug!("digest: {}", digest);

    let mut headers = String::new();
    headers.push_str(format!("(request-target): {} {}\n", method, path.display()).as_str());
    headers.push_str(format!("host: {}\n", host).as_str());
    headers.push_str(format!("date: {}\n", date).as_str());
    headers.push_str(format!("digest: {}", digest).as_str());

    debug!("signed_string: {:?}", headers);

    let signature = key.sign(headers.as_bytes());

    let header = format!(
        "keyId=\"{}\",algorithm=\"rsa-sha256\",headers=\"(request-target) host date digest\",signature=\"{}\"",
        key.key_id, signature
    );

    debug!("header: {}", header);

    HttpSignature {
        signature: header,
        digest,
        date,
    }
}

pub fn signed_post(
    key: Key,
    target @ Target { host, path }: Target,
    mut extra_headers: HeaderMap,
    body: &[u8],
) -> (StatusCode, Vec<u8>) {
    let HttpSignature {
        signature,
        digest,
        date,
    } = sign_request(key, target, Method::POST, body);

    let url = format!("https://{}{}", host, path.display());

    let mut req = ureq::post(url);
    {
        let headers = req.headers_mut().unwrap();
        headers.extend(extra_headers.drain());
    }
    let response = req
        .header("Signature", signature)
        .header("host", host)
        .header("date", date)
        .header("digest", digest)
        .send(body)
        .expect("Failed to send request");

    (
        response.status(),
        response
            .into_body()
            .read_to_vec()
            .expect("Failed to read the response as bytes"),
    )
}
